*! Nikolas Mittag 27dec2017

cap program drop cdtest
program define cdtest, rclass
version 12.0

syntax varlist(min=2 numeric) [if] [in], SAMple(varlist numeric min=1 max=1) [Weights(varlist numeric min=2 max=2)] [BAndwidth0(string)] [BAndwidth1(string)] [BSIter(integer 1000)] [Blocksize(integer 1000)] [Discrete(varlist)] [Treatmenteffect] [Commonsupport]

*temporary variables and names
tempname pval bs ckstat te tmp tmp2 K x_cont bw1 bw0 a
tempvar w0 w1 d0 d1 tmp_var
*variable lists
gettoken y X : varlist
local X=strtrim("`X'")
local `x_cont': list X - discrete
if strmatch("`bandwidth0'","")==1 {
	mat def `bw0'=J(1,1,.)
}
else mat def `bw0'=`bandwidth0'
if strmatch("`bandwidth1'","")==1 {
	mat def `bw1'=J(1,1,.)
}
else mat def `bw1'=`bandwidth1'


*only keep required variables and observations
preserve
marksample touse
qui keep if `touse'
qui drop if missing(`sample')==1
if strmatch("`weights'","")==0 {
	foreach x of varlist `weights' {
		qui drop if missing(`x')
	}
}
keep `varlist' `sample' `weights'

*restrict to common support 
****not sure this is ideal
if strmatch("`commonsupport'","")==0 {
	foreach x of varlist `discrete' {
		qui tab `x' if `sample'==1, matrow(`tmp')
		qui tab `x' if `sample'==0, matrow(`tmp2')
		local i=1
		mat def `a'=1
		capture noisily mat def `a'=(`tmp'-`tmp2')'*(`tmp'-`tmp2')
		while (`a'[1,1]!=0) {
			if `tmp'[`i',1]!=`tmp2'[`i',1] {
				qui drop if `x'==min(`tmp'[`i',1],`tmp2'[`i',1])
				qui tab `x' if `sample'==1, matrow(`tmp')
				qui tab `x' if `sample'==0, matrow(`tmp2')
			}
			local i=`i'+1
			capture mat def `a'=(`tmp'-`tmp2')'*(`tmp'-`tmp2')
		}
	}
	*continuous variables
	foreach x of varlist ``x_cont'' {
		qui sum `x' if `sample'==1
		local `tmp'=`r(min)'
		local `tmp2'=`r(max)'
		qui sum `x' if `sample'==0
		qui drop if `x'<max(`r(min)',``tmp'')
		qui drop if `x'>min(`r(max)',``tmp2'')
	}
}

**setup
*sample sizes
local N=_N
sum `sample', meanonly
local n1=`r(sum)'
local n0=`N'-`n1'
*other
local `te'=1-strmatch("`treatmenteffect'","")

sort `sample'

***call mata routine for test stat and p-value
if strmatch("`weights'","")==0 {
	gettoken w0 w1: weights
	mata: `ckstat'=ck_stat("`y'","``x_cont''","`discrete'","`w0'","`w1'","`bw0'","`bw1'",`n0',`n1',`N',`bsiter',``te'',`blocksize')
}
else mata: `ckstat'=ck_stat("`y'","``x_cont''","`discrete'","1","1","`bw0'","`bw1'",`n0',`n1',`N',`bsiter',``te'',`blocksize')
mata: st_numscalar("`ckstat'",`ckstat')

*program returns: r(ck_stat), r(p), r(bs) r(bsiter), r(N), r(n0), r(n1), r(y), r(X)
return scalar p=`pval'
return matrix bs_dist=`bs'
return scalar bsiter=`bsiter'
return scalar N=`N'
return scalar n0=`n0'
return scalar n1=`n1'
return scalar ck_stat=`ckstat'
return local y="`y'"
return local X="`X'"
mata: mata clear

***display results
****add number of excluded observations if restricted to common support?
di "Conditional Kolmogorov test whether the distribution of `y' given `X' differs by `sample'"
di "Test statistic: " `ckstat'
di "p-value: " `pval'
di "Number of observations: `N' (=`n0'+`n1')" 

restore 

end

***Mata routine: ck statistic and p-value
*data needs to be sorted by t, and may only contain estimation sample
mata
real scalar ck_stat(string scalar y, string rowvector cont, string rowvector disc, string scalar ws0, string scalar ws1, string scalar bw0, string scalar bw1, real scalar n0, real scalar n1, real scalar N, real scalar bsiter, real scalar te, real scalar BlockSize) {
/*Add missing definitions. Likely don't need to define as many variables. Should clean it up in the end.*/
real scalar i, K, Kdisc, Kcont, Const0, Const1, Lambda0, Lambda1, Match0, Match1, MatchC0, MatchC1
real matrix IBs0, IBs1, Xd, IsSame0, IsSame1, Mdisc0, Mdisc1, r01, r10, s
real colvector w0, w1, gx, sd0, sd1, HDiag0, HDiag1, HDiagInvHalf0, HDiagInvHalf1, ub

/****weights. I just copied this, we can probably clean it up a bit. Would likely run a bit faster if we just define views for "discrete" and "continuous" instead of subsetting X. Need to figure out how to align bandwidth and variable locations? Get rid of X, X0, X1?****/
if (ws1=="1") {
	st_view(discrete0=.,(1,n0),tokens(disc))
	st_view(continuous0=.,(1,n0),tokens(cont))
	st_view(discrete1=.,(n0+1,N),tokens(disc))
	st_view(continuous1=.,(n0+1,N),tokens(cont))
	st_view(X=.,.,(tokens(cont),tokens(disc)))
	Bandwidth0=st_matrix(bw0)
	Bandwidth1=st_matrix(bw1)
	K=cols(X)
	Kdisc=cols(discrete0)
	Kcont=K-Kdisc
	/*Not sure how well this works if there are no discrete/continuous variables. May need to be fixed. Continuous=select((1..K),(rowsum(J(1,Kdisc,(1..K)'):==discrete):==0)'), Update: Hopefully fixed, see below */
	Continuous=select((1..K),(1..K):<=Kcont)
	Discrete=select((1..K),(1..K):>Kcont)
	if ((cols(Bandwidth0)==K)*(cols(Bandwidth1)==K)*(Bandwidth0!=.)*(Bandwidth1!=.)) {
	HDiag0=(Bandwidth0[Continuous]):^2
	HDiag1=(Bandwidth1[Continuous]):^2
	} else {
	sd0=(colsum((continuous0:-(colsum(continuous0)/n0)):^2)/(n0-1)):^(1/2)
	sd1=(colsum((continuous1:-(colsum(continuous1)/n1)):^2)/(n1-1)):^(1/2)
	HDiag0=((4/(Kcont+2))^(1/(Kcont+4))*n0^(-1/(Kcont+4))*sd0):^2
	HDiag1=((4/(Kcont+2))^(1/(Kcont+4))*n1^(-1/(Kcont+4))*sd1):^2
	}
	HDiagInvHalf0=(1:/HDiag0)'/2
	HDiagInvHalf1=(1:/HDiag1)'/2
	Const0=1/n0*(2*pi())^(-Kcont/2)*(det(diag(HDiag0)))^(-1/2)
	Const1=1/n1*(2*pi())^(-Kcont/2)*(det(diag(HDiag1)))^(-1/2)

	Xd=J(N,2,.)
	if (cols(discrete0)==0) {
	for (i=1; i<=N; i++) {
	Xd[i,]=(
	sum(exp(-((continuous0:-X[i,]):^2)*HDiagInvHalf0)),
	sum(exp(-((continuous1:-X[i,]):^2)*HDiagInvHalf1))
	)
	}
	Xd=Xd:*(Const0,Const1)
	} else {
	if ((cols(Bandwidth0)==K)*(cols(Bandwidth1)==K)*(Bandwidth0!=.)*(Bandwidth1!=.)) {
	Lambda0=Bandwidth0[Discrete]
	Lambda1=Bandwidth1[Discrete]
	} else {
	Lambda0=J(1,Kdisc,n0^(-2/(Kcont+4)))
	Lambda1=J(1,Kdisc,n1^(-2/(Kcont+4)))
	}
	Match0=1:-Lambda0
	Match1=1:-Lambda1
	MatchC0=J(1,Kdisc,.)
	MatchC1=J(1,Kdisc,.)
	for (k=1; k<=Kdisc; k++) {
	MatchC0[k]=Lambda0[k]/(rows(uniqrows(discrete0[,k]))-1)
	MatchC1[k]=Lambda1[k]/(rows(uniqrows(discrete1[,k]))-1)
	}
	IsSame0=J(n0,Kdisc,.)
	IsSame1=J(n1,Kdisc,.)
	Mdisc0=J(n0,2*Kdisc,.)
	Mdisc1=J(n1,2*Kdisc,.)
	for (i=1; i<=N; i++) {
	IsSame0[,]=discrete0:==X[i,Discrete]
	IsSame1[,]=discrete1:==X[i,Discrete]
	Mdisc0[,]=(Match0:^IsSame0,MatchC0:^(1:-IsSame0))
	Mdisc1[,]=(Match1:^IsSame1,MatchC1:^(1:-IsSame1))
	mdisc0=Mdisc0[,1]
	mdisc1=Mdisc1[,1]
	for (k=2; k<=(2*Kdisc); k++) {
	mdisc0=mdisc0:*Mdisc0[,k]
	mdisc1=mdisc1:*Mdisc1[,k]
	}
	Xd[i,]=(
	sum(exp(-((continuous0:-X[i,Continuous]):^2)*HDiagInvHalf0):*mdisc0),
	sum(exp(-((continuous1:-X[i,Continuous]):^2)*HDiagInvHalf1):*mdisc1)
	)
	}
	Xd=Xd:*(Const0,Const1)
	}
	r01=Xd[,1]:/Xd[,2]
	r10=Xd[,2]:/Xd[,1]
	w0=(J(n0,1,1/N)\r01[(n0+1)..N]*(n1/((n0+n1)*sum(r01[(n0+1)..N]))))
	w1=(r10[1..n0]*(n0/((n0+n1)*sum(r10[1..n0])))\J(n1,1,1/N))
}
else {
	st_view(w0=.,.,ws0)
	st_view(w1=.,.,ws1)
}
/*create view (containing y,x)*/
/****if views speed things up, then all that needs to be done is to generate variables of the right storage type for each large matrix and then define the relevant objects as in the lines below*/
st_view(s=.,.,(y, tokens(cont),tokens(disc)))


/***Computing sampling matrix***/
IBs0=((J(n0,1,1)\J(n1,1,0)),J(N,bsiter,.))
IBs1=((J(n0,1,0)\J(n1,1,1)),J(N,bsiter,.))
/****double check whether this works. maybe this can be done quicker by doing it simultaneously or at least with less assignments****/
for (i=2;i<=cols(IBs0);i++) {
    ub = (mm_colrunsum(w0),J(N,2,0))
    ub[,1] = ub[,1]/ub[rows(ub),1]
	ub = sort(ub\(uniform(n0,1),J(n0,2,1)),1)
	ub[,3]=mm_colrunsum(ub[,2])
	ub=sort(ub,(2,3))[1..N,]
	ub[2..N,3]=ub[2..N,3]-ub[1..N-1,3]	
	IBs0[,i]=ub[,3] 

    ub = (mm_colrunsum(w1),J(N,2,0))
    ub[,1] = ub[,1]/ub[rows(ub),1]
    ub = sort(ub\(uniform(n1,1),J(n1,2,1)),1)
	ub[,3]=mm_colrunsum(ub[,2])
	ub=sort(ub,(2,3))[1..N,]
	ub[2..N,3]=ub[2..N,3]-ub[1..N-1,3]	
	IBs1[,i]=ub[,3] 
}

/***Computing test statistics v02 ***/
BlockIndices=floor(range(1,N,BlockSize))
BlockIndices=BlockIndices+(J((rows(BlockIndices)-1),1,0)\1)
BlockMaxCK=J(rows(BlockIndices)-1,bsiter+1,.)
if (te) {
fy0=(IBs0:>0):*(colsum(IBs0:*s[,1])/n0) /*Here is generally loop over bootstrap iteration which fill predicted values of obseravations resampled to s0*/
fy1=(IBs1:>0):*(colsum(IBs1:*s[,1])/n1) /*Here is generally loop over bootstrap iteration which fill predicted values of obseravations resampled to s1*/
ry=(s[,1]:-fy0)\((s[,1]:-fy1)) 
for(bi=1; bi<rows(BlockIndices); bi++){
    BlockRows=BlockIndices[bi]..(BlockIndices[bi+1]-1)
	BlockN=cols(BlockRows)
	gx=J(BlockN,N,1)
	for(k=2;k<=(K+1); k++){
        gx=gx:*(J(BlockN,1,s[,k]'):<=s[BlockRows,k])
    }
	gxyIBs0=J(BlockN*2,bsiter+1,.)
	gxyIBs1=J(BlockN*2,bsiter+1,.)
	ggy=J(BlockN*2,N*2,.)
	for(b=1;b<(bsiter+2);b++){
	    ggy=J(BlockN*2,1,ry[,b]'):<=ry[(BlockRows,N:+BlockRows),b]
		gxyIBs0[,b]=((gx\gx):*ggy[,1..N])*IBs0[,b]
		gxyIBs1[,b]=((gx\gx):*ggy[,(N+1)..(N*2)])*IBs1[,b]
	}
	gxIBs0=gx*IBs0
	gxIBs1=gx*IBs1
	BlockMaxCK[bi,]=colmax(abs(gxyIBs0:*(gxIBs1\gxIBs1)-gxyIBs1:*(gxIBs0\gxIBs0)):*((IBs0[BlockRows,]:>0)\(IBs1[BlockRows,]:>0)))
}
} else {
for(bi=1; bi<rows(BlockIndices); bi++){
	BlockRows=BlockIndices[bi]..(BlockIndices[bi+1]-1)
	BlockN=cols(BlockRows)
	gx=J(BlockN,N,1)
	for(k=2;k<=(K+1); k++){
        gx=gx:*(J(BlockN,1,s[,k]'):<=s[BlockRows,k])
    }
	gy=(J(BlockN,1,s[,1]'):<=s[BlockRows,1])
	gxy=gx:*gy
	BlockMaxCK[bi,]=colmax(abs((gxy*IBs0):*(gx*IBs1)-(gxy*IBs1):*(gx*IBs0)):*((IBs0[BlockRows,]+IBs1[BlockRows,]):>0))
}
}
MaxCK=colmax(BlockMaxCK)/sqrt(n0*n1)
st_matrix(st_local("bs"),MaxCK[2..(bsiter+1)])
st_numscalar(st_local("pval"),sum(MaxCK[2..(bsiter+1)]:>MaxCK[1])/bsiter)
return(MaxCK[1])
}

end
